/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 * @file functions.h
 * @brief The header for base classes of function objects and several
 *        useful functions that are not supported in the standard
 *        template library (STL).
 * @author Kiminori MATSUZAKI, Yoshiki AKASHI
 */

#ifndef __FUNCTIONS_H__
#define __FUNCTIONS_H__

#include <utility>
#include <functional>
#include <climits>
#include <cfloat>

namespace sketo {
  namespace functions {
    /**
     * @brief A common class used in defining function objects.
     *
     * The idea of using template parameter of the form base<int(double)>,
     * for example, is from the implementation of boost::functions.
     */
    template <typename T> class base;

    template <typename R>
    class base <R (void)> {
    public:
      typedef R result_type;
    };
    
    template <typename R, typename A1>
    class base <R (A1)> {
    public:
      typedef R result_type;
      typedef A1 argument_type;
      typedef A1 arg1_type;
    };
    
    template <typename R, typename A1, typename A2>
    class base <R (A1, A2)> {
    public:
      typedef R result_type;
      typedef A1 first_argument_type;
      typedef A2 second_argument_type;
      typedef A1 arg1_type;
      typedef A2 arg2_type;
    };
    
    template <typename R, typename A1, typename A2, typename A3>
    class base <R (A1, A2, A3)> {
    public:
      typedef R result_type;
      typedef A1 type1;
      typedef A2 type2;
      typedef A3 type3;
    };
  }
}

#define SKETO_DEF_UNOP(name, rettype, intype, body)                                             \
  class name##_t : public sketo::functions::base <rettype (intype)> {                           \
  public:                                                                                       \
      rettype operator()(const intype &x) const {                                               \
        body                                                                                    \
      }                                                                                         \
  } name

#define SKETO_DEF_BINOP(name, rettype, intype1, intype2, body)                                  \
  class name##_t : public sketo::functions::base <rettype (intype1, intype2)> {                 \
  public:                                                                                       \
      rettype operator()(const intype1 &x, const intype2 &y) const {                            \
        body                                                                                    \
      }                                                                                         \
  } name

#define SKETO_DEF_TEROP(name, rettype, intype1, intype2, intype3, body)                         \
  class name##_t : public sketo::functions::base <rettype (intype1, intype2, intype3) > {       \
  public:                                                                                       \
      rettype operator()(const intype1 &x, const intype2 &y, const intype3 &z) const {          \
        body                                                                                    \
      }                                                                                         \
  } name


////////////////////////////////////////////////////////////////////////
// function objects
namespace sketo {
  namespace functions {
    template <typename A>
    struct identity : public base <A (A)> {
      A operator()(A x) const { return x; }
    };

    template <typename A, typename B>
    struct caster : public base <B (A)> {
      B operator()(A x) const { return static_cast<B>(x); }
    };
	
    template <typename A, typename B>
    struct fst : public base <A (std::pair<A, B>)> {
      A operator()(std::pair <A, B> x) const { return x.first; }
    };
    
    template <typename A, typename B>
    struct snd : public base <B (std::pair<A, B>)> {
      B operator()(std::pair <A, B> x) const { return x.second; }
    };

    template <typename A, typename B>
    struct mkpair : public base <std::pair<A, B> (A, B)> {
      std::pair<A, B> operator()(const A &a, const B &b) const { return std::pair<A, B>(a, b); }
    };
    
    template <typename A, typename B>
    struct left : public base <A (A, B)> {
      A operator()(A x, B) const { return x; }
    };
    
    template <typename A, typename B>
    struct right : public base <B (A, B)> {
      B operator()(A, B y) const { return y; }
    };

    template <typename A>
    struct square : public base <A (A)> {
      A operator()(A x) const { return x * x; }
    };

    template <typename A>
    struct max : public base <A (A, A)> {
      A operator()(A x, A y) const { return (x < y) ? y : x; }
    };

    template <typename A>
    struct min : public base <A (A, A)> {
      A operator()(A x, A y) const { return (x < y) ? x : y; }
    };

    template <typename OP>
    struct swap_adapter_t : public base<typename OP::result_type (typename OP::second_argument_type, typename OP::first_argument_type)> {
      OP op;
      swap_adapter_t(const OP &op) : op(op) { }
      typename OP::result_type operator()(typename OP::second_argument_type x, typename OP::first_argument_type y) const {
	return op(y, x);
      }
    };

    template <typename OP>
    swap_adapter_t<OP>
    swap(const OP &op) {
      return swap_adapter_t<OP>(op);
    }

    template <typename F, typename G>
    struct composition : public base <typename F::result_type (typename G::argument_type)> {
      const F f;
      const G g;
      composition(const composition<F, G> &fg) : f(fg.f), g(fg.g) { }
      const composition<F, G> & operator=(const composition<F, G> &fg) {
        (F)f = fg.f;
        (G)g = fg.g;
        return *this;
      }
      composition(const F&f, const G&g) : f(f), g(g) { }
      typename F::result_type operator()(const typename G::argument_type &x) const { return f(g(x)); }
    };

  }
}

template <typename A>
inline A
identity_element(std::plus<A>)
{
  return A(0);
}

template <typename A>
inline A
identity_element(std::multiplies<A>)
{
  return A(1);
}

template <typename A>
inline A
identity_element(sketo::functions::max<A>);

template <>
inline double
identity_element<double>(sketo::functions::max<double>)
{
  return -DBL_MAX;
}

template <>
inline int
identity_element(sketo::functions::max<int>)
{
  return INT_MIN;
}

template <typename A>
inline A
identity_element(sketo::functions::min<A>);

template <>
inline double
identity_element<double>(sketo::functions::min<double>)
{
  return DBL_MAX;
}

template <>
inline int
identity_element(sketo::functions::min<int>)
{
  return INT_MAX;
}


#endif // __FUNCTIONS_H__

