/*** * **** * ***** ********* ** ****** ***** *** ***** ********
**
** Copyright (c) 2005-2013, SkeTo Project
** All rights reserved.
**
** This file is part of the SkeTo Library. This library is free
** software; you can redistribute it and/or modify it under the
** terms of the MIT License (see COPYING in the package).
** 
** This library is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
**
*** * **** * ***** ********* ** ****** ***** *** ***** ********/

/**
 *
 * @file matrix_skeletons_predef.h
 * @brief A set of predefinition of class names, etc.
 */
#ifndef __MATRIX_SKELETONS_PREDEF_H__
#define __MATRIX_SKELETONS_PREDEF_H__

namespace sketo {
  namespace _impl {
    namespace ms_impl_select {
      struct ReduceSimple {
        template <typename OPLUS, typename OTIMES, typename AS, typename A>
        static A
        reduce(const OPLUS &op, const OTIMES &ot, const AS &as);
      };
      struct ReduceDC {
        template <typename OPLUS, typename OTIMES, typename AS, typename A>
        static A
        reduce(const OPLUS &op, const OTIMES &ot, const AS &as);
      };
    }
  }
  namespace matrix_skeletons {
    typedef _impl::ms_impl_select::ReduceDC ReduceDC;
    typedef _impl::ms_impl_select::ReduceSimple ReduceSimple;
  };
  namespace _impl {
    struct shift_block_info {
      mindex rank;             // target rank
      msize size;              // size of the block
      mindex begin;            // local begin index of the block
      mindex target_begin;     // target begin index of the block
      bool from_outside;       // whether is block is comming from outside  
      mindex global_begin;     // original global begin index of the block
      int tag;                 // tag used for communication
      shift_block_info() : rank(-1, -1), size(0,0), begin(-1,-1), target_begin(-1, -1), from_outside(false) { }
    };

    std::ostream& operator<<(std::ostream &os, const shift_block_info& sbi) {
      os << "b=" << sbi.begin << ", g=" << sbi.global_begin << ", t=" << sbi.target_begin << ", s=" << sbi.size << ", r=" << sbi.rank << ", o=" << sbi.from_outside << "tag=" << sbi.tag;
      return os;
    }

    template <typename A>
    struct shift_info_buf {
      A *bufs[4];
      int ref;
      shift_info_buf() : ref(0) {
        for(int i = 0; i < 4; i++) bufs[i] = NULL;
        inc_ref();
      } 
      void inc_ref() { ref++; }
      void dec_ref() { ref--; if(ref==0) { free_buf(); } }
      void free_buf() {
        for(int i = 0; i < 4; i++) if(bufs[i] != NULL) delete [] bufs[i];
      }
      A* reserve(int i, int len) {
        if(bufs[i]==NULL) bufs[i] = new A[len];
        return bufs[i];
      }
      A* get_buf(int i) { return bufs[i]; }
    };
    template <int di, int dj, typename A>
    struct shift_info {
      shift_block_info to_sbis[4];
      shift_block_info from_sbis[4];

      shift_info_buf<A> *buf;
      
      const msize pa;
      const mindex rs;
      const mindexsize mis;
      const msize global_size;
      bool ready;
      shift_info(const shift_info &o) : pa(o.pa), rs(o.rs), mis(o.mis), global_size(o.global_size), ready(o.ready) {
        for(int i = 0; i < 4; i++) {
          to_sbis[i] = o.to_sbis[i];
        }
        for(int i = 0; i < 4; i++) {
          from_sbis[i] = o.from_sbis[i];
        }
        buf = o.buf;
        buf->inc_ref();
      }
      shift_info(msize global_size) :
        pa(_impl::dist_matrix_buffer<A>::compute_process_arrangement(global_size, sketo::procs)),
        rs(_impl::dist_matrix_buffer<A>::compute_ranks(pa, sketo::rank)),
        mis(_impl::dist_matrix_buffer<A>::compute_local_array(global_size, pa, rs)), 
        global_size(global_size),
        ready(false) {
        compute_info_4(mindex(di, dj), to_sbis);
        compute_info_4(mindex(-di, -dj), from_sbis);
        buf = new shift_info_buf<A>();        
      }
      ~shift_info() {
        buf->dec_ref();
      }
      mindex self_begin;
      mindex self_index;
      mindex self_end;
      void gen_index() {
        for(int i = 0; i < 4; i++) {
          if(!need_receive_buf(i) && from_sbis[i].rank.i != -1) {
            self_begin = from_sbis[i].begin;
            self_end = from_sbis[i].begin + from_sbis[i].size;
            self_index = from_sbis[i].target_begin;
           }
        }
        
      }
      template <typename AS>
      A local_get(const mindex i, const AS &as) { // too slow.
        if((i.i >= self_begin.i) && (i.j >= self_begin.j) && (i.i < self_end.i) && (i.j < self_end.j)) {
          return as.local_get(i-self_begin + self_index);
        }
        
        for(int k = 0; k < 4; k++) {
          if(need_receive_buf(k)) {
            const int ii = i.i - from_sbis[k].begin.i; 
            const int jj = i.j - from_sbis[k].begin.j;
            if(ii >= 0 && jj >= 0 && ii < from_sbis[k].size.i && jj < from_sbis[k].size.j) {
              return (buf->get_buf(k))[ii * from_sbis[k].size.j + jj];
            }
          }
        }
        assert(false);
        return A();
      }
      bool need_send_communication(int i) const {
        if(to_sbis[i].rank.i == -1) return false;
        if(rs == to_sbis[i].rank) return false;
        return true;
      }
      bool need_receive_buf(int i) const {
        if(from_sbis[i].rank.i == -1) return false;
        if(rs == from_sbis[i].rank) return from_sbis[i].from_outside;
        return true;
      }
      bool need_receive_communication(int i) const {
        return need_receive_buf(i) && rs != from_sbis[i].rank;
      }
      bool is_self_move(int i) const {
        return rs == from_sbis[i].rank && !from_sbis[i].from_outside;
      }
      bool is_ready() {
        bool old = ready;
        ready = true;
        return old;
      }
      void compute_info_4(mindex d, shift_block_info *sbis) {
        sbis[0] = compute_info(d, mindex(0, 0));

        // full-column transfer?
        const bool fullcol = sbis[0].size.i == mis.size.i;

        // full-row transfer?
        const bool fullrow = sbis[0].size.j == mis.size.j;

        if(fullcol) {
          if(fullrow) {
            sbis[1] = shift_block_info(); // no use
            sbis[2] = shift_block_info(); // no use
            sbis[3] = shift_block_info(); // no use
          } else {
            sbis[1] = compute_info(d, mindex(0, mis.size.j-1));
            sbis[2] = shift_block_info(); // no use
            sbis[3] = shift_block_info(); // no use
          }
        } else {
          if(fullrow) {
            sbis[1] = shift_block_info(); // no use
            sbis[2] = compute_info(d, mindex(mis.size.i-1, 0));
            sbis[3] = shift_block_info(); // no use
          } else {
            sbis[1] = compute_info(d, mindex(0, mis.size.j-1));
            sbis[2] = compute_info(d, mindex(mis.size.i-1, 0));
            sbis[3] = compute_info(d, mindex(mis.size.i-1, mis.size.j-1));
          }
        }        
      }
        
      shift_block_info compute_info(mindex d, mindex index)
      {
        // global index of the local point (corner) shifted by d
        mindex index_ul = mis.index + index + d;
        
        // the rank of the globa lindex and its local index in the rank
        std::pair<mindex, std::pair<mindex, bool> > ri_ul = _impl::dist_matrix_buffer<A>::compute_rank_from_index(global_size, sketo::procs, index_ul, pa);
        
        // get information about the rank
        mindexsize mis_ul = _impl::dist_matrix_buffer<A>::compute_local_array(global_size, pa, ri_ul.first);
        shift_block_info sbi;
        sbi.rank  = ri_ul.first;
        sbi.from_outside = ri_ul.second.second;
        
        mindex upleft = mindex(0,0);
        
        // compute the size of the shift block
        if(index.i == 0) {
          if(index.j == 0) {
            sbi.size  = msize(mis_ul.size.i - ri_ul.second.first.i, mis_ul.size.j - ri_ul.second.first.j);
            sbi.global_begin = index_ul;
          } else {
            sbi.size  = msize(mis_ul.size.i - ri_ul.second.first.i, ri_ul.second.first.j + 1);
            upleft = mindex(0, -sbi.size.j+1);
          }
        } else {
          if(index.j == 0) {
            sbi.size  = msize(ri_ul.second.first.i + 1, mis_ul.size.j - ri_ul.second.first.j);
            upleft = mindex(-sbi.size.i+1, 0);
          } else {
            sbi.size  = msize(ri_ul.second.first.i + 1, ri_ul.second.first.j + 1);
            upleft = mindex(-sbi.size.i+1, -sbi.size.j+1);
          }
        }
        sbi.global_begin = index_ul + upleft;
        sbi.target_begin = ri_ul.second.first + upleft;
        sbi.begin = index + upleft;

        // sbi.size may be greater than mis.size by 1.
        if(sbi.size.i > mis.size.i) sbi.size = msize(mis.size.i, sbi.size.j);
        if(sbi.size.j > mis.size.j) sbi.size = msize(sbi.size.i, mis.size.j);
        
        const mindex tb = sbi.target_begin;
        if(tb.i == 0) {
          if(tb.j == 0) {  // upper-left corner
            sbi.tag = 0;
          } else {         // upper-right corner
            sbi.tag = 1;
          }
        } else {
          if(tb.j == 0) {  // lower-left corner
            sbi.tag = 2;
          } else {         // lower-right corner
            sbi.tag = 3;
          }
        }
        
        return sbi;
      }
    };
    
    class ms_adpt {
    public:
      
      template <typename A>
      static bool
      has_no_data(const dist_matrix<A> &a) {
        return a.data == NULL;
      }

      template <typename A, typename OP, typename OT>
      static void
      global_reduce(const OP &oplus,
                    const OT &otimes,
                    A* pval,
                    msize pa);

      template <typename OPLUS, typename OTIMES, typename A, typename AS>
      static dist_matrix<A>
      scan(dist_matrix<A>* bs,
           const OPLUS &oplus,
           const OTIMES &otimes,
           const AS &as);
        
      template <typename OPLUS, typename OTIMES, typename A, typename AS>
      static dist_matrix<A>
      scanr(dist_matrix<A>* bs,
            const OPLUS &oplus,
            const OTIMES &otimes,
            const AS &as);
      
      template <typename OPLUS, typename OTIMES, typename AS, typename A, typename IINDEXER, typename JINDEXER>
      static dist_matrix<A>
      scan_common(dist_matrix<A> *bs,
                  const OPLUS &oplus,
                  const OTIMES &otimes,
                  const AS &as,
                  const IINDEXER& iindexer,
                  const JINDEXER& jindexer);
      
      template <typename A, typename OP, typename INDEXER>
      static void
      global_scan_common(const OP& oplus,
                         const int n,
                         A* pfirsts,
                         A* ptotals,
                         const msize pa,
                         const mindex rs,
                         const INDEXER &genidnex);
      
   
      template <int di, int dj, typename F, typename AS, typename A>
      static void
      shift_common(dist_matrix<A> *bs, shift_info<di, dj, A> &si, const F&f, const AS&as);
      
#ifdef __SKETO_NO_FUSION__
      template <typename F, typename A>
      static dist_matrix<A>
      generate(msize global_size, const F& f);

      template <typename F, typename A, typename B>
      static dist_matrix<B>
      map(const F& f,
          const dist_matrix<A>& as);

      template <typename F, typename A, typename B>
      static dist_matrix<B>
      map_with_index(const F& f,
                     const dist_matrix<A>& as);
      
      template <typename F, typename A, typename B, typename C>
      static dist_matrix<C>
      zipwith(const F& f,
              const dist_matrix<A>& as,
              const dist_matrix<B>& bs);
      
#else
      template <typename OP, typename OT, typename AS, typename A>
      static void
      scanobj_subeval(dist_matrix<A>* bs, const OP op, const OT ot, const AS as);
      
      template <typename OP, typename OT, typename AS, typename A>
      static void
      scanrobj_subeval(dist_matrix<A>* bs, const OP op, const OT ot, const AS as);
      
      template <typename OP, typename OT, typename AS, typename A>
      static dist_matrix<A>*
      scanobj_ow_impl(dist_matrix<A>* bs, const OP op, const OT ot, const AS as);
      
      template <typename OP, typename OT, typename AS, typename A>
      static dist_matrix<A>*
      scanrobj_ow_impl(dist_matrix<A>* bs, const OP op, const OT ot, const AS as);
      
#endif

    };

  }

}




#endif // __MATRIX_SKELETONS_PREDEF_H__
