#include <iostream>
#include <cmath>
#include <utility>

#include <sketo/list_skeletons.h>
#ifndef GSHIFT_DUMMY
#include "gshift.h"
#else
#include "gshift_dummy.h"
#endif

namespace sl = sketo::list_skeletons;

#include "testcase.h"

struct array1D {
  double data[imax+2];
};

array1D zero_1D;

struct plus_1D_t : public sketo::functions::base<array1D (array1D, array1D)> {
  array1D operator()(const array1D &x, const array1D &y) const {
    array1D ret;
    for (int i = 0; i < imax+2; i++) {
      ret.data[i] = x.data[i] + y.data[i];
    }
    return ret;
  }
} plus_1D;

struct times_1D_t : public sketo::functions::base<array1D (array1D, array1D)> {
  array1D operator()(const array1D &x, const array1D &y) const {
    array1D ret;
    for (int i = 0; i < imax+2; i++) {
      ret.data[i] = x.data[i] * y.data[i];
    }
    return ret;
  }
} times_1D;

struct times_scalar_1D : public sketo::functions::base<array1D (array1D)> {
  double scalar;
  times_scalar_1D(double scalar) : scalar(scalar) { }
  array1D operator()(const array1D &x) const {
    array1D ret;
    for (int i = 0; i < imax+2; i++) {
      ret.data[i] = scalar * x.data[i];
    }
    return ret;
  }
};

struct minus_1D_t : public sketo::functions::base<array1D (array1D, array1D)> {
  array1D operator()(const array1D &x, const array1D &y) const {
    array1D ret;
    for (int i = 0; i < imax+2; i++) {
      ret.data[i] = x.data[i] - y.data[i];
    }
    return ret;
  }
} minus_1D;

struct sqsum_1D_t : public sketo::functions::base<double (array1D)> {
  double operator()(const array1D &x) const {
    double ret = 0.0;
    for (int i = 1; i <= imax; i++) {
      ret += (x.data[i] * x.data[i]);
    }
    return ret;
  }
} sqsum_1D;

struct sum_1D_t : public sketo::functions::base<double (array1D)> {
  double operator()(const array1D &x) const {
    double ret = 0.0;
    for (int i = 1; i <= imax; i++) {
      ret += x.data[i];
    }
    return ret;
  }
} sum_1D;




struct gen_b_t : public sketo::functions::base<array1D (int)> {
  array1D operator()(int jk) const {
    const int j = jk % (jmax + 2); jk /= (jmax + 2);
    const int k = jk;
    array1D ret;
    for (int i = 0; i < imax + 2; i++) {
      double v;
      if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1) ||
	  (k == 0) || (k == kmax + 1)) {
	  v = 0.0;
      } else {
	v = 2.0;
	if ((i > 1) && (i < imax)) v += 2.0; else v += 1.0;
	if ((j > 1) && (j < jmax)) v += 2.0; else v += 1.0;
	if ((k > 1) && (k < kmax)) v += 2.0; else v += 1.0;
      }
      ret.data[i] = v;
    }
    return ret;
  }
} gen_b;

struct get_f_t : public sketo::functions::base<array1D (int)> {
  array1D operator()(int jk) const {
    const int j = jk % (jmax + 2); jk /= (jmax + 2);
    const int k = jk;
    array1D ret;
    for (int i = 0; i < imax + 2; i++) {
      if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1) ||
	  (k == 0) || (k == kmax + 1)) {
	ret.data[i] = 0.0;
      } else {
	ret.data[i] = 1.0;
      }
    }
    return ret;
  }
} gen_f;
    
void make_vector(double *ap, double *ae, double *aw, double *an, double *as, double *ab, double *at,
		 sketo::dist_list<array1D> *b, sketo::dist_list<array1D> *f)
{
  *b = sl::generate((kmax + 2) * (jmax + 2), gen_b);
  *f = sl::generate((kmax + 2) * (jmax + 2), gen_f);

  *ap = 2.0;
  *ae = *aw = *an = *as = *at = *ab = 1.0;
}

struct local_1D : public sketo::functions::base<array1D (array1D)> {
  double ap, aw, ae;
  local_1D(double ap, double aw, double ae) : ap(ap), aw(aw), ae(ae) { }
  array1D operator()(const array1D &x) const {
    array1D ret;
    for (int i = 1; i <= imax; i++) {
      ret.data[i] = ap * x.data[i] + aw * x.data[i-1] + ae * x.data[i+1];
    }
    return ret;
  }
};

void probmv(double ap, double ae, double aw, double an, double as, double at, double ab,
	    const sketo::dist_list<array1D> &b, sketo::dist_list<array1D> *c,
	    const sketo::dist_list<array1D> &f)
{
  *c = sl::zipwith(times_1D, f,
       sl::zipwith(plus_1D, sl::map(local_1D(ap, aw, ae), b),
       sl::zipwith(plus_1D, sl::map(times_scalar_1D(as), sl::shiftl(zero_1D, b)),
       sl::zipwith(plus_1D, sl::map(times_scalar_1D(an), sl::shiftr(zero_1D, b)),
       sl::zipwith(plus_1D, sl::map(times_scalar_1D(ab), gshift(  jmax+2, zero_1D, b)),
		            sl::map(times_scalar_1D(at), gshift(-(jmax+2), zero_1D, b)))))));
}

std::pair<int, double>
bicgstab_band(double ap, double ae, double aw ,double an, double as, double at, double ab,
	      sketo::dist_list<array1D> *x,
	      sketo::dist_list<array1D> *b,
	      const sketo::dist_list<array1D> &f)
{
//   double er0 = EPS;

  sketo::dist_list<array1D> c, r, p, r0, e, y, v;

  probmv(ap, ae, aw, an, as, at, ab, *x, &c, f);

  r = sl::zipwith(minus_1D, *b, c);
  double c1 = sl::reduce(std::plus<double>(),
			 sl::map(sqsum_1D, r));

//   if (c1 < er0) {
//     return std::pair<int, double>(0, c1);
//   }

  p = r.clone();
  r0 = r.clone();

  std::pair<int, double> ret;
  for (int it = 0; it < ITRMAX; it++) {

    probmv(ap, ae, aw, an, as, at, ab, p, &y, f);

    double c2 = sl::reduce(std::plus<double>(),
			   sl::map(sum_1D, 
				   sl::zipwith(times_1D, r0, y)));
    double alp = c1 / c2;

    e = sl::zipwith(minus_1D,
		    r,
		    sl::map(times_scalar_1D(alp), y));

    probmv(ap, ae, aw, an, as, at, ab, e, &v, f);

    double ev = sl::reduce(std::plus<double>(),
			   sl::map(sum_1D, 
				   sl::zipwith(times_1D, e, v)));
    double vv = sl::reduce(std::plus<double>(),
			   sl::map(sum_1D, 
				   sl::zipwith(times_1D, v, v)));
    
    double c3 = ev / vv;
    *x = sl::zipwith(plus_1D,
		     *x, sl::zipwith(plus_1D,
				     sl::map(times_scalar_1D(alp), p),
				     sl::map(times_scalar_1D(c3), e)));
    r = sl::zipwith(minus_1D, e, sl::map(times_scalar_1D(c3), v));
    
    double rr = sl::reduce(std::plus<double>(),
			   sl::map(sum_1D, 
				   sl::zipwith(times_1D, r, r)));
//     if (std::fabs(rr) < er0) { ret = std::pair<int, double>(it + 1, rr); break; }

    c1 = sl::reduce(std::plus<double>(),
		    sl::map(sum_1D,
			    sl::zipwith(times_1D, r0, r)));
    double bet = c1/(c2 * c3);
    p = sl::zipwith(plus_1D,
		    r, sl::zipwith(plus_1D,
				   sl::map(times_scalar_1D(bet), p),
				   sl::map(times_scalar_1D(-bet * c3), y)));
    
    ret = std::pair<int, double>(it + 1, rr);
  }
  return ret;
}

struct gen_x_t : public sketo::functions::base<array1D (int)> {
  array1D operator()(int jk) const {
    const int j = jk % (jmax + 2); jk /= (jmax + 2);
    const int k = jk;
    array1D ret;
    for (int i = 0; i < imax+2; i++) {
      if ((i == 0) || (i == imax + 1) || (j == 0) || (j == jmax + 1) || (k == 0) || (k == kmax + 1)) {
	ret.data[i] = 0.0;
      } else {
	ret.data[i] = 1.0 + 0.01;
      }
    }
    return ret;
  }
} gen_x;

int sketo::main(int, char**)
{
  // zero_1D
  for (int i = 0; i < imax+2; i++) {
    zero_1D.data[i] = 0.0;
  }
  
  double ap, ae, aw, an, as, at, ab;
  sketo::dist_list<array1D> b, f;
  make_vector(&ap, &ae, &aw, &an, &as, &at, &ab, &b, &f);

  sketo::dist_list<array1D> x = sl::generate((kmax + 2) * (jmax + 2), gen_x);
  
  double st, et;
  MPI_Barrier(MPI_COMM_WORLD);
  st = MPI_Wtime();
#ifndef PROBMV_ONLY
  std::pair<int, double> result = bicgstab_band(ap, ae, aw, an, as, at, ab,
						&x, &b, f);
  std::cout << "error = " << result.second << std::endl;
#else
  ap = -4.5;
  sketo::dist_list<array1D> y((kmax + 2) * (jmax + 2));
  // avoid too-much optimization
  double tmpval = 0.0;
  for (int i = 0; i < ITRMAX * 2; i++) {
    probmv(ap, ae, aw, an, as, at, ab, x, &y, f);
    tmpval += sketo::_impl::dist_list_buffer<array1D>::get_buffer(y)->buf[0].data[0];
  }
  sketo::cout << tmpval << std::endl;
#endif
  MPI_Barrier(MPI_COMM_WORLD);
  et = MPI_Wtime();
  sketo::cout << "time = " << (et - st) << std::endl;
  return 0;
}



